#ifndef __MODEL_MGR_H__
#define __MODEL_MGR_H__

#include <stdint.h>
#include "cmsis_os2.h"
#include "ipc.h"             /* for MULTI_MODEL_MAX */
#include "base.h"
#include "model_type.h"
//#include "common.h"
#include "kdp_flash_table.h"

/* a term means ALL_MODELS */
#define KMDW_MODEL_ALL_MODELS  -1

/* MAX model count for DME and flash */
#define KMDW_MODEL_MAX_MODEL_COUNT MULTI_MODEL_MAX

/* These crc values will ONLY change if model is changed in FLASH */
#define MODEL_INFO_SUM32_VALUE  0x00eee074

#define MODEL_IN_DDR            1
#define MODEL_IN_FLASH          0

enum kmdw_model_rc {
    // 0 - 9 is reserved for ncpu return
    // defined in ipc.h
    // IMAGE_STATE_INACTIVE == 0
    // IMAGE_STATE_ACTIVE == 1
    // IMAGE_STATE_NPU_DONE == 2
    // IMAGE_STATE_DONE == 3
    KMDW_MODEL_RUN_RC_ABORT = 10,
    KMDW_MODEL_RUN_RC_ERROR = 11,
    KMDW_MODEL_RUN_RC_END
};

typedef struct kmdw_model_img_run_time_s {
    uint32_t    round_trip_time;
    uint32_t    pre_proc_time;
    uint32_t    npu_proc_time;
    uint32_t    post_proc_time;
} kmdw_model_run_time_t;

//a mask for easy read fw_info data
typedef struct kmdw_model_fw_info_s {
    uint32_t                model_count;
    struct kdp_model_s      models[1];   // just maintain the address of dynamical count of models
} kmdw_model_fw_info_t;

//a mask for easy read fw_info data
typedef struct kmdw_model_fw_info_ext_s {
    uint32_t                model_dram_addr_end;
    uint32_t                model_total_size;
    uint32_t                model_checksum;
} kmdw_model_fw_info_ext_t;


/* ############################
 * ##    Public Functions    ##
 * ############################ */

/**
 * @brief Init model functionality
 */
void kmdw_model_init(void);

/**
 * @brief A wrapper of load_model from flash
 * @param model_index_p: model info index,
 *                    0-n: info_index of model to load
 *                    -1 means to load all models
 * @return 0: failes;  1: OK(means 1 model is loaded)
 */
int32_t kmdw_model_load_model(int8_t model_info_index_p);


/**
 * @brief A wrapper of load_model_info
 * @param [in] is_model_from_ddr: if model is from ddr/host command
 * @return reloaded model count; 0 means failed
 */
int32_t kmdw_model_reload_model_info(bool from_ddr);


/**
 * @brief Refresh all models
 * @return refreshed model count; 0 means failed
 */
int32_t kmdw_model_refresh_models(void);


/**
 * @brief Output model_info of specified index
 * @param[in] idx_p the index of programmed models
 * @return model_info defined in ipc.h
 */
struct kdp_model_s* kmdw_model_get_model_info(int idx_p);


/**
 * @brief Output total number of models and all model IDs
 * @param[in] from_ddr: if models are in ddr or flash
 * @param[in] trust_ddr_data trust ddr data
 * @return a list of model count, id0, id1, id2
 * @note about trust_ddr_data, ddr data might be dirty after soft/hw reset
 *       trust_ddr_data means no need to check validness of ddr data
 */
uint32_t *kmdw_model_get_all_model_info(bool from_ddr, bool trust_ddr_data);

/**
 * @brief Output model end address defined in model file
 * @param[in] from_ddr if models are in ddr or flash
 * @param[in] trust_ddr_data trust ddr data
 * @return model end address when loaded
 * @note about trust_ddr_data, ddr data might be dirty after soft/hw reset
 *       trust_ddr_data means no need to check validness of ddr data
 */
uint32_t kmdw_model_get_model_end_addr(bool from_ddr, bool trust_ddr_data);

/**
 * @brief Output crc value
 * @param[in] from_ddr: if models are in ddr or flash
 * @param[in] trust_ddr_data trust ddr data
 * @return crc value
 * @note about trust_ddr_data, ddr data might be dirty after soft/hw reset
 *       trust_ddr_data means no need to check validness of ddr data
 */
uint32_t kmdw_model_get_crc(bool from_ddr, bool trust_ddr_data);

/**
 * @brief Get the buffer address of fw_info
 * @return the buffer address of fw_info
 */
kmdw_model_fw_info_t* kmdw_model_get_fw_info_buf(void);


/**
 * @brief Check if a model is in flash
 * @param[in] model_type model type
 * @return 1 - the model is in flash; 0 - not in flash
 */
int kmdw_model_is_model_in_flash(uint32_t model_type);


/**
 * @brief Specify output address for model run in ncpu/npu
 * @return always 0
 * @note must be called after kmdw_model_config_model()
 */
int32_t kmdw_model_config_result(osEventFlagsId_t result_evt, uint32_t result_evt_flag);


/**
 * @brief Config model image
 * @param img_cfg image config
 * @param crop_box image crop config
 * @param pad_values image padding config
 * @param ext_param extra param
 */
void kmdw_model_config_img(struct kdp_img_cfg *img_cfg, struct kdp_crop_box_s *crop_box,
                           struct kdp_pad_value_s *pad_values, void *ext_param);


/**
 * @brief Get raw image config
 * @param idx image index
 * @return raw image config
 */
struct kdp_img_raw_s* kmdw_model_get_raw_img(int idx);


/**
 * @brief Run model
 * @param tag model tag
 * @param output model output
 * @param model_type model type
 * @param dme DME mode
 * @return kmdw_model_rc
 */
int kmdw_model_run(const char *tag, void *output, uint32_t model_type, bool dme);


/**
 * @brief Abort model execution
 */
void kmdw_model_abort(void);


/**
 * @brief Get round-trip/pre/npu/post times
 * @param [in]  img_idx raw image index
 * @param [out] run_time run time of pre/npu/post/
 * @return void
 */
void kmdw_model_get_run_time(int img_idx, kmdw_model_run_time_t *run_time);


/**
 * @brief Check if a model is loaded in memory
 * @param[in] model_type model type
 * @return 1 - loaded in memory; 0 - not loaded
 */
int kmdw_model_is_model_loaded(uint32_t model_type);

/**
 * @brief Set model storage location
 * @param[in] model_inddr // model is from flash : false, ddr : true
 */
void kmdw_model_set_location(bool model_inddr);

/**
 * @brief Get model storage location
 * @return 1 - DDR; 0 - Flash
 */
bool kmdw_model_get_location(void);

#ifdef EMBED_CMP_NPU
/**
 * @brief A wrapper of adding or update specific model configuration
 * @param model_type model type
 * @param cmd_len model command length
 * @param wt_len model weight length
 * @param input_len input data length
 * @param output_len output data length
 * @param setup_len setup data length
 * @param cmd_mem_addr model command memory address
 * @param wt_mem_addr model weight memory address
 * @param input_mem_addr input data memory address
 * @param output_mem_addr output data memory address
 * @param setup_mem_addr setup data memory address
 * @return int8_t  0: config ok
 *                -1: model not found
 */
int8_t kmdw_model_add_update_model(uint32_t model_type, 
                                   int cmd_len, int wt_len, int input_len, int output_len, int setup_len,
                                   uint32_t cmd_mem_addr, uint32_t wt_mem_addr, 
                                   uint32_t input_mem_addr, uint32_t output_mem_addr, uint32_t setup_mem_addr);

#endif // EMBED_CMP_NPU

#if DEBUG

/**
 * @brief Dump model debug info
 */
void kmdw_model_dump_model_info(void);

#endif

#endif
