// #define ENABLE_DBG_LOG

#include <string.h>
#include "cmsis_os2.h"
#include "power_manager.h"

#include "kmdw_console.h"
#include "kmdw_memory.h"
#include "kdrv_gdma.h"
#include "kdrv_power.h"
#include "kdrv_scu_ext.h"
#include "kdp_system.h"

#include "usbd_hal.h"

#include "model_type.h"

#include "buffer_object.h"

#include "kmdw_inference_client.h"
#include "kdp2_ipc_cmd.h"
#include "kdp2_inf_app_yolo.h"

#include "hico_simulator.h"

#ifdef ENABLE_DBG_LOG
#define dbg_log(__format__, ...) kmdw_printf("[kp hico sim]"__format__, ##__VA_ARGS__)
#else
#define dbg_log(__format__, ...)
#endif

#define DEFAULT_IMAGE_BUF_SIZE (1 * 1024 * 1024) // image + descriptor
#define DEFAULT_RESULT_COUNT 4
#define DEFAULT_RESULT_BUF_SIZE (1 * 1024 * 1024) // result + descriptor

#define JTAG_MAGIC_ADDRESS 0x10100000
#define KDP2_BOOT_CONFIG_ADDRESS 0x10100100

#define JTAG_MAGIC_VALUE 0xFEDCBA01
#define BOOT_FROM_FLASH 0xA

typedef struct
{
    uint32_t boot_type;     // 0xA = flash-boot, others = usb-boot
    uint8_t loader_ver[4];  // fw loader version numbers
    uint8_t scpu_fw_ver[4]; // SCPU fw version numbers
    uint8_t ncpu_fw_ver[4]; // NCPU fw version numbers
} kdp2_boot_config_t;

static osThreadId_t hico_sim_tid = 0;

// simulate receiced image buffers from the image sensor
static buffer_object_t g_image_bobjs[TOTAL_IMAGE_NUM];

// record raw image size
static uint32_t g_image_size = 0;

extern int kdp2_cmd_handler_initialize(void);
extern int kdp2_cmd_handle_kp_command(uint32_t command_header_buf); // for new KDP2 commands
extern int kdp2_cmd_handle_legend_kdp_command(uint32_t command_buffer); // for old KDP commands

// vendor-specific control transfer setup packet notify
bool usb_user_control_callback(usbd_hal_setup_packet_t *setup)
{
    bool ret = false;

    dbg_log("control bRequest = 0x%x\n", setup->bRequest);

    switch (setup->bRequest)
    {
    case KDP2_CONTROL_REBOOT:
    {
        dbg_log("control reboot\n");
        kdrv_power_sw_reset();
        break;
    }
    case KDP2_CONTROL_SHUTDOWN:
    {
        dbg_log("control shutdown\n");
        power_manager_shutdown();
        break;
    }
    case KDP2_CONTROL_FIFOQ_RESET:
    {
        dbg_log("control fifoq reset\n");
        usbd_hal_terminate_all_endpoint();
        ret = true;
        break;
    }

    default:
        ret = false;
        break;
    }

    return ret;
}

// this thread receive the command from host SW
void kdp2_hico_simulator_usb_cmd_thread(void *arg)
{
    dbg_log("[%s] start !\n", __FUNCTION__);

#define CMD_BUF_SIZE (1 * 1024 * 1024)

    // a big enough buffer to receive an image
    uint32_t cmd_buf = kmdw_ddr_reserve(CMD_BUF_SIZE);

    while (1)
    {
        uint32_t txLen = CMD_BUF_SIZE;
        kdrv_status_t usb_sts = usbd_hal_bulk_receive(KDP2_USB_ENDPOINT_DATA_OUT, (void *)cmd_buf, &txLen, osWaitForever);
        if (usb_sts != KDRV_STATUS_OK) // KDRV_STATUS_USBD_TRANSFER_TERMINATED or KDRV_STATUS_USBD_TRANSFER_DISCONNECTED
        {
            dbg_log("[%s] bulk receive is terminated, sts %d\n", __FUNCTION__, usb_sts);
            continue;
        }

        dbg_log("[%s] usb recv addr 0x%x len %d\n", __FUNCTION__, (void *)cmd_buf, txLen);

        kp_inference_header_stamp_t *header_stamp = (kp_inference_header_stamp_t *)cmd_buf;

        if (header_stamp->magic_type == KDP2_MAGIC_TYPE_COMMAND) // standard KDP2 commands
        {
            // borrow image fifo to handle KDP2 commands as well
            dbg_log("[%s] handle kdp2 command = 0x%x\n", __FUNCTION__, header_stamp->job_id);
            // handle kdp2 commands ...
            kdp2_cmd_handle_kp_command(cmd_buf);
            continue;
        }
        else if (header_stamp->magic_type == KDP2_MAGIC_TYPE_CUSTOMIZED) // customized commands
        {
            int status = KP_SUCCESS;

            // customized command to load image to DDR
            if (header_stamp->job_id == JOB_ID_LOAD_IMAGE)
            {
                dbg_log("[%s] load image from PLUS to fifoQ\n", __FUNCTION__);

                load_image_command_t *load_img_params = (load_image_command_t *)cmd_buf;

                dbg_log("[%s] image_index: %d\n", __FUNCTION__, load_img_params->index);

                g_image_size = IMAGE_WIDTH * IMAGE_WIDTH * 2;

                if (load_img_params->index >= TOTAL_IMAGE_NUM)
                    status = KP_ERROR_INVALID_PARAM_12;
                else
                {
                    dbg_log("[%s] free buffer address: 0x%x\n", __FUNCTION__, g_image_bobjs[load_img_params->index].buffer_addr);

                    // construct a header for app yolo inference
                    kdp2_ipc_app_yolo_inf_header_t *inf_header = (kdp2_ipc_app_yolo_inf_header_t *)g_image_bobjs[load_img_params->index].buffer_addr;
                    inf_header->header_stamp.magic_type = KDP2_MAGIC_TYPE_INFERENCE;
                    inf_header->header_stamp.total_size = sizeof(kdp2_ipc_app_yolo_inf_header_t) + g_image_size;
                    inf_header->header_stamp.job_id = KDP2_INF_ID_APP_YOLO;
                    inf_header->header_stamp.status_code = KP_SUCCESS; // FIXME

                    inf_header->inf_number = 0;
                    inf_header->width = IMAGE_WIDTH;
                    inf_header->height = IMAGE_HEIGHT;
                    inf_header->channel = 3;
                    inf_header->model_id = TINY_YOLO_V3_224_224_3;
                    inf_header->image_format = KP_IMAGE_FORMAT_RGB565;
                    inf_header->model_normalize = KP_NORMALIZE_KNERON;

                    // copy raw image buffer to image fifo queue
                    kdrv_gdma_memcpy(g_image_bobjs[load_img_params->index].buffer_addr + sizeof(kdp2_ipc_app_yolo_inf_header_t), cmd_buf + sizeof(load_image_command_t), g_image_size);
                }

                // send return code of this customized command to host SW
                usbd_hal_bulk_send(KDP2_USB_ENDPOINT_DATA_IN, (void *)&status, sizeof(status), 1000);

                // if all images are loaded (save to image fifo queue), start hico simulation thread
                if (load_img_params->index == TOTAL_IMAGE_NUM - 1)
                    osThreadFlagsSet(hico_sim_tid, 0x100000);
            }
            else
            {
                kmdw_printf("[%s] !!!!!! wrong job_id = %d\n", __FUNCTION__, header_stamp->job_id);
            }
        }
        else if ((header_stamp->magic_type & 0xFFFF) == KDP_MSG_HDR_CMD) // very speical case for old arch. fw update
        {
            // handle legendary kdp commands, should be as few as possible
            dbg_log("handle legendary kdp command = 0x%x\n", header_stamp->job_id);
            kdp2_cmd_handle_legend_kdp_command(cmd_buf);
        }
        else
        {
            kmdw_printf("[%s] error ! buffer begin with incorrect magic_type 0x%x, txLen %d\n", __FUNCTION__, header_stamp->magic_type, txLen);
        }
    }
}

// this thread sends inference result to host
void kdp2_hico_simulator_usb_result_thread(void *arg)
{
    dbg_log("[%s] start !\n", __FUNCTION__);

    while (1)
    {
        uint32_t result_buf_addr;
        int result_buf_length;

        // get result data from result fifo queue with blocking wait
        kmdw_inference_client_result_request(&result_buf_addr, &result_buf_length, osWaitForever);

        // then send inference result
        kp_inference_header_stamp_t *header_stamp = (kp_inference_header_stamp_t *)result_buf_addr;

        header_stamp->job_id = JOB_ID_THIS_IS_INF_RESULT; // for now it tells host SW that this is an inference result

        dbg_log("[%s] send hico simulation inference result via usb, addr 0x%x len %d\n", __FUNCTION__, (void *)result_buf_addr, header_stamp->total_size);

        // send result to the host SW, blocking wait
        kdrv_status_t usb_sts = usbd_hal_bulk_send(KDP2_USB_ENDPOINT_DATA_IN, (void *)result_buf_addr, header_stamp->total_size, osWaitForever);
        if (usb_sts != KDRV_STATUS_OK) // KDRV_STATUS_USBD_TRANSFER_TERMINATED or KDRV_STATUS_USBD_TRANSFER_DISCONNECTED
        {
            kmdw_printf("XXXX send hico simulation inference result via usb failed, sts = %d\n", usb_sts);
        }

        // return free buf back to queue
        kmdw_inference_client_result_put_free_buffer(result_buf_addr, result_buf_length, osWaitForever);
    }
}

// this thead simulates sending raw image from image sensor to host SW and also enqueuing it to image fifo queue
void kdp2_hico_simulator_thread(void *arg)
{
    dbg_log("[%s] start !\n", __FUNCTION__);

    osThreadFlagsWait(0x100000, osFlagsWaitAny, osWaitForever);

    int i = 0;

    while (1)
    {
        // suppose the received raw image is at g_image_bobjs[i].buffer_addr + sizeof(kdp2_ipc_app_yolo_inf_header_t)
        // note that we need to attach a header stamp before raw image buffer
        kp_inference_header_stamp_t *header_stamp = (kp_inference_header_stamp_t *)g_image_bobjs[i].buffer_addr;

        // send raw image to host SW
        header_stamp->job_id = JOB_ID_THIS_IS_IMAGE;

        uint32_t transfer_length = sizeof(kdp2_ipc_app_yolo_inf_header_t) + g_image_size;

        dbg_log("[%s] send hico simulation image via usb, index %d, addr 0x%x len %d\n", __FUNCTION__, i, (void *)g_image_bobjs[i].buffer_addr, transfer_length);

        kdrv_status_t usb_sts = usbd_hal_bulk_send(KDP2_USB_ENDPOINT_DATA_IN, (void *)g_image_bobjs[i].buffer_addr, transfer_length, osWaitForever);

        if (usb_sts != KDRV_STATUS_OK) // KDRV_STATUS_USBD_TRANSFER_TERMINATED or KDRV_STATUS_USBD_TRANSFER_DISCONNECTED
        {
            kmdw_printf("XXXX send hico simulation image via usb failed, sts = %d\n", usb_sts);
            continue; // FIXME: skip this frame if sending to host timeout, actually timeout should not occur
        }

        // enqueue raw image to image fifo queue
        header_stamp->job_id = KDP2_INF_ID_APP_YOLO; // for now it tells inference core to do app yolo inference on the raw image

        buffer_object_t bobj;

        // get free buffer from image fifo queue
        kmdw_inference_client_image_get_free_buffer(&bobj.buffer_addr, &bobj.length, osWaitForever, false);

        dbg_log("[%s] free buffer address: 0x%x\n", __FUNCTION__, bobj.buffer_addr);

        // enqueue raw image buffer to image fifo queue
        // you can copy received data from the image sensor to this address
        // note that we need to attach a header stamp before raw image buffer
        // kdrv_gdma_memcpy(bobj.buffer_addr + sizeof(kdp2_ipc_app_yolo_inf_header_t), g_image_bobjs[i].buffer_addr + sizeof(kdp2_ipc_app_yolo_inf_header_t), g_image_size, NULL, NULL);
        // we don't copy here because they are already the same, just for demo purpose and efficiency

        kmdw_inference_client_image_enqueue(bobj.buffer_addr, bobj.length, osWaitForever, false);

        i++;

        // reset index i
        if (i == TOTAL_IMAGE_NUM)
            i = 0;

        // put some delay here to simulate the output frame rate of image sensor
        // osDelay(50);
    }
}

////////////////////////////////////////////////////////////
#define RECOVERY_MARK_POS (SdRAM_MEM_BASE + SdRAM_MEM_SIZE - 64)

// KDP2 Inference Interface for HICO, simulation code
// image input
// image + inference output
int kdp2_hico_simulator_init()
{
    // retrieve real serial number here from efuse
    // then convert it to hex string format

    uint32_t uid = 0;

    uid = kdp_sys_get_kn_number();

    int32_t sidx = 0;
    uint8_t kn_num_string[32] = {0};
    for (int i = 7; i >= 0; i--)
    {
        uint32_t hex = (uid >> i * 4) & 0xF;
        kn_num_string[sidx] = (hex < 10) ? '0' + hex : 'A' + (hex - 10);
        sidx += 2;
    }

    // HICO Mode
    uint16_t bcdDevice = KP_KDP2_FW_HICO_MODE;

    if (*((uint32_t *)JTAG_MAGIC_ADDRESS) == JTAG_MAGIC_VALUE)
    {
        kmdw_printf("FW is running in JTAG mode\n");
        bcdDevice |= KP_KDP2_FW_JTAG_TYPE;
    }
    else
    {
        kdp2_boot_config_t *bConfig = (kdp2_boot_config_t *)KDP2_BOOT_CONFIG_ADDRESS;
        if (bConfig->boot_type == BOOT_FROM_FLASH)
        {
            kmdw_printf("KDP2 FW is running in flash-boot mode\n");
            bcdDevice |= KP_KDP2_FW_FLASH_TYPE;

            kmdw_printf("boot ncpu fw from flash\n");
            SCU_EXTREG_CM4_NCPU_CTRL_SET_wakeup(1); // run ncpu
        }
        else
        {
            kmdw_printf("KDP2 FW is running in usb-boot mode\n");
            bcdDevice |= KP_KDP2_FW_USB_TYPE;
        }
    }

    // this is about recovery mode
    *(uint32_t *)RECOVERY_MARK_POS = 0;

    usbd_hal_initialize(kn_num_string, bcdDevice, usb_user_control_callback);
    usbd_hal_set_enable(true);

    kdrv_gdma_initialize();

    /* Allocate memory for image buffers */
    uint32_t image_buf_begin_addr = kmdw_ddr_reserve(TOTAL_IMAGE_NUM * DEFAULT_IMAGE_BUF_SIZE);
    for (uint32_t i = 0; i < TOTAL_IMAGE_NUM; i++)
    {
        g_image_bobjs[i].buffer_addr = image_buf_begin_addr + DEFAULT_IMAGE_BUF_SIZE * i;
        g_image_bobjs[i].length = DEFAULT_IMAGE_BUF_SIZE;
    }

    /* Allocate memory for result buffers */
    uint32_t result_buf_begin_addr = kmdw_ddr_reserve(DEFAULT_RESULT_COUNT * DEFAULT_RESULT_BUF_SIZE);
    buffer_object_t result_bobjs[DEFAULT_RESULT_COUNT];
    for (uint32_t i = 0; i < DEFAULT_RESULT_COUNT; i++)
    {
        result_bobjs[i].buffer_addr = result_buf_begin_addr + DEFAULT_RESULT_BUF_SIZE * i;
        result_bobjs[i].length = DEFAULT_RESULT_BUF_SIZE;
    }

    // Prepare free image buffers
    for (int i = 0; i < TOTAL_IMAGE_NUM; i++)
        kmdw_inference_client_image_put_free_buffer(g_image_bobjs[i].buffer_addr, g_image_bobjs[i].length, osWaitForever);

    // Prepare free result buffers
    for (int i = 0; i < DEFAULT_RESULT_COUNT; i++)
        kmdw_inference_client_result_put_free_buffer(result_bobjs[i].buffer_addr, result_bobjs[i].length, osWaitForever);

    osThreadId_t image_thread_id = osThreadNew(kdp2_hico_simulator_usb_cmd_thread, NULL, NULL);
    osThreadId_t result_thread_id = osThreadNew(kdp2_hico_simulator_usb_result_thread, NULL, NULL);
    hico_sim_tid = osThreadNew(kdp2_hico_simulator_thread, NULL, NULL);

    if (image_thread_id == NULL || result_thread_id == NULL || hico_sim_tid == NULL)
    {
        kmdw_printf("kdp2 usb companion create threads failed\n");
        return -1;
    }

    // wow ! fifoq can also handle command
    kdp2_cmd_handler_initialize();

    return 0;
}
